/*
 *
 * Copyright (c) 2007 Apple Inc.
 * All Rights Reserved
 * 
 * This file is part of Hourglass.
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation is hereby granted without fee, provided that the
 * above copyright notice and this permission/disclaimer notice is
 * retained in all copies or modified versions, and that both notices
 * appear in supporting documentation.  THE COPYRIGHT HOLDERS PROVIDE
 * THIS SOFTWARE "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE COPYRIGHT
 * HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 * RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu
 * any improvements that they make and grant redistribution rights to the
 * University of Utah.
 * 
 * Author: Ryan Du Bois (rdub@apple.com)
 */

#include <mach/mach.h>
#include "hourglass.h"

#undef DEBUG
/* darwin has this, but it always returns unimplemented */
#undef HAVE_MLOCKALL	

#ifdef DEBUG
const inline char *k_err(const kern_return_t code);

const inline char *k_err(const kern_return_t code)
{
	switch(code)
	{
	case KERN_INVALID_POLICY: return "KERN_INVALID_POLICY";
	case KERN_INVALID_ARGUMENT: return "KERN_INVALID_ARGUMENT";
	case KERN_SUCCESS: return "KERN_SUCCESS";
	default: return "unknown";
	}
}
#endif

/* I'm defining my own type
 * since all the policy base structures
 * we care about start with a common field, 
 * which also happens to be the only one we
 * care about: **base_priority**
 *
 * Hopefully this simplifies things for us
 */
typedef struct {
	int base_priority;
}fake_policy_base;

static void setup_policy(policy_t policy, policy_base_t *pbase, fake_policy_base **ptr,
			unsigned int *size)
{
	switch(policy)
	{
		case POLICY_TIMESHARE:
			*pbase = calloc(1, sizeof(struct policy_timeshare_base));
			*size = sizeof(struct policy_timeshare_base)/sizeof(int);
			break;
		case POLICY_RR:
			*pbase = calloc(1, sizeof(struct policy_rr_base));
			*size = sizeof(struct policy_rr_base)/sizeof(int);
			break;
	}
	if(!*pbase)
	{
		perror("Could not allocate policy_base structure");
		exit(-1);
	}
	*ptr = (fake_policy_base *) *pbase;
	#ifdef DEBUG
	printf("exiting setup_policy(): p: %08x, param: %08x\n", (unsigned int) pbase, 
								(unsigned int) ptr);
	#endif
}

static int get_max_rr_prio()
{
	kern_return_t err = 0;
	struct policy_rr_info rr;
	mach_msg_type_number_t count = sizeof(struct policy_rr_info)/sizeof(int);

	if((err = thread_info(mach_thread_self(), THREAD_SCHED_RR_INFO,
			(thread_info_t) &rr, &count)) != KERN_SUCCESS)
	{
		printf("thread_info: could not get max RR priority\n");
		#ifdef DEBUG
		printf("error: %d: %s\n", err, k_err(err));
		#endif
		exit(err);
	}
	DBGPrint (8, ("max_realtime_priority == %d\n", rr.max_priority));
	return rr.max_priority;
}


static unsigned int make_parameters(enum thrd_pri_t pri, policy_t *policy,
								policy_base_t *p,
								int max_realtime_pri,
								int *niceval)
{
	fake_policy_base *param;
	unsigned int size = 0;
	
	switch (pri) {
		
		case IDLE:
			setup_policy(POLICY_TIMESHARE, p, &param, &size);
			param->base_priority = 0;
			*policy = POLICY_TIMESHARE;
			*niceval = 19;
			break;
			
		case LOW:
			setup_policy(POLICY_TIMESHARE, p, &param, &size);
			param->base_priority = 0;
			*policy = POLICY_TIMESHARE;
			*niceval = 10;
			break;
			
		case NORMAL:
			setup_policy(POLICY_TIMESHARE, p, &param, &size);
			param->base_priority = 0;
			*policy = POLICY_TIMESHARE;
			*niceval = 0;
			break;
			
		case HIGH:
			setup_policy(POLICY_TIMESHARE, p, &param, &size);
			param->base_priority = 0;
			*policy = POLICY_TIMESHARE;
			*niceval = -10;
			break;
			
		case HIGHEST:
			setup_policy(POLICY_TIMESHARE, p, &param, &size);
			param->base_priority = 0;
			*policy = POLICY_TIMESHARE;
			*niceval = -19;
			break;
			
		case RTLOW:
			setup_policy(POLICY_RR, p, &param, &size);
			param->base_priority = max_realtime_pri / 4;
			*policy = POLICY_RR;
			break;
			
		case RTMED:
			setup_policy(POLICY_RR, p, &param, &size);
			param->base_priority = max_realtime_pri / 2;
			*policy = POLICY_RR;
			break;
			
		case RTHIGH:
			setup_policy(POLICY_RR, p, &param, &size);
			param->base_priority = max_realtime_pri - 2;
			*policy = POLICY_RR;
			break;
			
		case TIMER_PRI:
			setup_policy(POLICY_RR, p, &param, &size);
			param->base_priority = max_realtime_pri - 1;
			*policy = POLICY_RR;
			break;
			
		case MAIN_PRI:
			setup_policy(POLICY_RR, p, &param, &size);
			param->base_priority = max_realtime_pri;
			*policy = POLICY_RR;
			break;
			
		default:
			printf ("Hrm: unexpected arg to set_priority: %d\n", pri);
			exit (-1);
	}
	
	DBGPrint (5, ("sched_priority = %d\n", param->base_priority));
	DBGPrint (5, ("policy = %d\n", *policy));
	
	#ifdef DEBUG
	printf("exiting make_parameters(): p: %08x, param: %08x, policy: %d\n", p, 
					param, *policy);
	#endif
	return size;
}

static int activate_scheduler(enum thrd_pri_t pri)
{
	struct policy_timeshare_base ts;
	struct policy_rr_base rr;
	policy_base_t ptr;
	kern_return_t err;
	policy_t pol;
	unsigned int count;
	unsigned char went_to_rr=0;
	
	switch (pri) {
		
		case IDLE:
		case LOW:
		case NORMAL:
		case HIGH:
		case HIGHEST:
			/* go into timesharing policy */
			ts.base_priority = 0;
			ptr = (policy_base_t) &ts;
			pol = POLICY_TIMESHARE;
			count = sizeof(struct policy_timeshare_base)/sizeof(int);
			went_to_rr = 0;
			break;
			
		case RTLOW:
		case RTMED:
		case RTHIGH:
		case TIMER_PRI:
		case MAIN_PRI:
			/* go into RR policy */
			rr.base_priority = 0;
			ptr = (policy_base_t) &rr;
			pol = POLICY_RR;
			count = sizeof(struct policy_rr_base)/sizeof(int);
			went_to_rr = 1;
			break;
			
		default:
			printf ("Hrm: unexpected arg to activate_target_priority: %d\n", pri);
			exit (-1);
	}
	
	/* set the right policy */
	if((err = thread_policy(mach_thread_self(), pol,
					ptr, count, 0)) != KERN_SUCCESS)
	{
		printf("thread_policy returned an error!\n");
		#ifdef DEBUG
		printf("error: %d: %s\n", err, k_err(err));
		#endif
		exit(err);
	}
	/* we are now in the correct policy */
	
	/* if we happen to be switching to RR, get the max priority now */
	if(went_to_rr)
		return get_max_rr_prio();
	
	/* let them know we're still in timesharing */
	return 0;
}


/* @fn set_priority(enum thrd_pri_t pri)
 * Setup the correct scheduler and priority values on a per-thread basis
 *
 * @param pri Thread Priority to setup
 */
void set_priority (enum thrd_pri_t pri)
{
	policy_base_t pbase;
	policy_t policy;
	int niceval = 0;
	int max_rr_prio = 0;
	unsigned int size;
	kern_return_t err;

	/* step 1: activate target scheduler */
	max_rr_prio = activate_scheduler(pri);
	
	/* step 2: build the correct priority parameters */
	size = make_parameters(pri, &policy, &pbase, max_rr_prio, &niceval);
	
	/* step all: debug */
	#ifdef DEBUG
	struct policy_rr_base *rr = (struct policy_rr_base *)pbase;
	struct policy_timeshare_base *ts = (struct policy_timeshare_base *)pbase;
	
	if(policy == POLICY_TIMESHARE)
	{
		printf("TIMESHARE ADJUSTMENT:\n");
		printf("\tbase_priority: %d\n", ts->base_priority);
		printf("\tnice: %d\n", niceval);
	}
	else if(policy == POLICY_RR)
	{
		printf("RR ADJUSTMENT:\n");
		printf("\tbase_priority: %d\n", rr->base_priority);
		printf("\tsize: %d (%s)\n", size, (size == POLICY_RR_BASE_COUNT) ? "ok" : "wrong");
	}
	else
		printf("problem: in set_priority(), while adjusting, wrong policy type\n");
	#endif
	
	/* step 3: adjust priority */
	if((err = thread_policy(mach_thread_self(), policy,
					pbase, size, 0)) != KERN_SUCCESS)
	{
		printf("thread_policy returned an error!\n");
		#ifdef DEBUG
		printf("in set_priority(): error: %d: %s\n", err, k_err(err));
		#endif
		exit(err);
	}
	
	/* step 3b: renice, if needed */
	if (policy == POLICY_TIMESHARE) {
		errno = 0;
		setpriority(PRIO_PROCESS, 0, niceval);
		if (errno != 0) 
		{
			perror ("setpriority");
			exit (errno);
		}
	}
	
	/* step 4: profit! */
}

#define MLOCK_ERR_1a "    **** WARNING: lack permissions to lock pages ****\n"
#define MLOCK_ERR_1b "    **** WARNING: mlockall() not supported on this platform ****\n"
#define MLOCK_ERR_2  "    **** results from this run may be unreliable ****\n"

void lock_pages (void)
{
#ifdef HAVE_MLOCKALL
	int res;
	
	res = mlockall (MCL_CURRENT | MCL_FUTURE);
	if (res != 0) {
		perror ("mlockall");
		printf ("%s%s", MLOCK_ERR_1a, MLOCK_ERR_2);
	}
#else
	kern_return_t err;
	
	printf ("%s", MLOCK_ERR_1b);
	printf("    **** trying Mach's thread_wire() instead ****\n");
		
	if((err = thread_wire(mach_host_self(), mach_thread_self(), 1)) != KERN_SUCCESS)
	{
		printf("    **** thread_wire returned an error ****\n");
        printf("    **** Try tunning as root ****\n");
		#ifdef DEBUG
		printf("in lock_pages(): error: %d: %s\n", err, k_err(err));
		#endif
		printf("%s%s", MLOCK_ERR_1a, MLOCK_ERR_2);
	}
    else
        printf("    **** Looks good... Continuing ****\n");
	
#endif
}

